/*
*    cwtmp: compact a wtmp file by discarding uninteresting entries.
*    1990-11-16 John H. DuBois III (john@armory.com)
*    1993-02-28 added entry discarding based on name prefix
*    1994-03-14 Modified error msg to reflect ambiguity of getutent() failure.
*    1994-06-22 Exit on error.  Expanded help.
*    1994-07-03 Process command line options.  Exit 0 at end of main().
*    1994-07-03 Warn about bad size of file, since this is a non-obvious thing
*               that makes getutent() fail.
*    1994-10-23 Fixed so that /etc/wtmp really is the default.
*    1994-12-14 Do not use utmp lib routines; poor error reporting & they
*               refuse to operate on a corrupt (wrong size) file.
*               Added all options.
*    1995-07-23 Added prsw options.
*               Fixed bug that caused coredump if a ut_type < 0 was returned.
*    1995-10-15 Added it options.
*    1995-12-30 Write help to stdout, not stderr.
*    1996-01-06 Began adding e option.
*    1997-01-01 Added ouzOUt options.
*    1997-04-30 Make -x take value.
*               Make -U also clear all DEAD_PROCESS entries.
*    1997-05-11 Re-port to 3.2v4.
*    1999-07-15 2.4.1 Do not write raw utmp records to a tty.
*    1999-08-28 2.4.2 Make -oz work again; broken by previous change.
*    2000-01-12 2.5 Added R option.
*    2001-11-12 3.0 Added e option.
*    2003-01-15 3.0.1 Use [UW]TMP[X]_FILE macros.
*    2005-05-07 3.1 Added N option and automatic type detection.
*    2008-03-04 3.2 Let "-" indicate standard input
*    2009-03-06 3.3 Added H option.
*    2010-10-24 3.4 Linux re-port.
*
*    todo:
*          Rewrite the safe-zero option so it make sense; it should really
*          overwrite *all* records after processing them.
*/

#include <time.h>
#include <stdio.h>
#include <pwd.h>
#include <fcntl.h>
#include <unistd.h>
#include "nutmp.h"
#include <sys/types.h>
#include <signal.h>
#include <sys/errno.h>
#if defined(M_UNIX) && !defined(_SCO_XPG_VERS)
#define ftruncate(fd,size) chsize(fd,size)
#endif

/* For Linux */
#include <stdlib.h>
#include <string.h>

char **StrToArr(char *S,char Sep[],char EscChar);

char SCCSstring[] = "@(#) cwtmp 3.3 2009-03-06";

/* This will be at least as large as a utmp record */
static struct utmpx emptyRec;

static int Debug = 0;
extern int errno;

main(argc,argv)
int argc;
char *argv[];
{
    void *orig_record;
    struct utmpx *ux;
    char *InFile;
    char *DeleteName = NULL;
    int DeleteLen = 0;
    static int RemoveTypes[UTMAXTYPE+1];
    int TypesGiven = 0;
    long Skip = 0;
    long Write = 0;
    long Value;
    int Readable = 0;
    int verboseRecords = 0;
    char **IDlist = NULL;
    char **id;
    int DelInitCmds = 0;
    int Extended = 0;		/* We are processing an extended utmpx file */
    int formatSpecified = 0;	/* Extended/not-extended format was specified on cmd line */
    int goodRec;
    int numWrite;
    /*
     * True if -z is given and not overwriting input.
     * Replace deleted records with zero records.
     */
    int zero = 0;
    /*
     * overwrite is true if any of oOuU are given.
     * Output is being written to same file as input.
     */
    int overwrite = 0;
    /*
     * safeZero is true if the input file is being overwritten and records
     * are to be cleared.  This is done when records are discarded by emptying
     * them, and also when a utmp file is being compacted in place and "safe
     * operation" has been called for; in this case emptying the records
     * immediately after copying them prevents records from being duplicated if
     * processing is aborted.
     */
    int safeZero = 0;
    /*
     * True if overwriting and doing zero-replacement.
     * If overZero is true, safeZero is also true.
     * safeZero accomplishes the zero-overwriting,
     * so overZero is used only to suppress various other actions.
     */
    int overZero = 0;
    int outFD = 1;
    /* emptyFD is a duplicate of the input file, for use in writing empty
       utmp records */
    int emptyFD;
    unsigned writeRec = 0;
    int recordsSkipped = 0;
    int tell = 0;
    int uflag = 0;
    int autoClean = 0;
    unsigned zeroWritePos; /* Position to do a zero record write */
    unsigned emptyFDpos = 0; /* Current position of emptyFD */
    size_t ut_size = sizeof(struct utmp);

    int Err = 0;
    char c;			/* Option character */
    extern char *optarg;	/* For optargs */
    extern int optind;		/* For optargs */
    int terrno;


    while ((c = getopt(argc,argv,"ad:efhHi:nNoOrRp:w:s:tTuUvx:z0123456789")) != -1)
	switch (c) {
	case 'h':
	    printf(
"cwtmp: compress wtmp file data by removing uninteresting entries.\n\
Usage: cwtmp [-afhHnNrtTvxzoOuU] [-<type>] [-d<prefix>] [-i<ID[:ID...]>]\n\
             [-[psw]<numrec>] [file]\n\
cwtmp compresses a wtmp file by discarding the uninteresting utmp entries\n\
that make up a large fraction of the typical wtmp file.  If no filename is\n\
given, %s is read.  utmp entries in the file are written to the\n\
output, except entries of type LOGIN_PROCESS, INIT_PROCESS, and EMPTY,\n\
which are skipped.\n\
If the input file is not a multiple of the size of a utmp record and one\n\
of the v, a, or f flags is not given, cwtmp will exit with an error\n\
message and nonzero status.\n\
Options:\n\
-h: Print this help.\n"
#ifdef EXTENDED
"-e: Force processing of the named files as extended utmp files (e.g. utmpx,\n\
    wtmpx).  This changes the default filename to %s (or if -u or -U\n\
    is given, %s).\n\
-N: Force processing of the named files as non-extended utmp files.\n\
If neither -e nor -N is given, cwtmp attempts to automatically determine the\n\
    types of input files.\n"
#endif
"-v: Check every utmp record for validity.  If an invalid record is found,\n\
    scan the input character by character attempting to find the start of\n\
    the next utmp record.  Invalid records are not output.\n\
-a: Autoverify.  If the input file has a size that is not a multiple of\n\
    the size of a utmp record, the -v option is turned on.\n\
-x<debug-level>: Turn on debugging at the given level.  At level 1, a\n\
    description of which wtmp file verification tests fail is printed.\n\
    At higher levels, more verbose diagnostics are printed.\n\
-r: Print records in readable form, rather that writing the utmp structs.\n\
-H: When records are printed in a readable form, use a more human-readable\n\
    format: the record types and times are expanded.\n\
-R: Short for -rH.\n\
-t: Tell what records would be discarded or emptied but do no other output.\n\
    Use -H as well to get a more verbose display.\n\
-z: Write an empty (zeroed) record in place of any deleted records.  Empty\n\
    records are not written in place of bad sections skipped due to\n\
    verification checks.  If -w is given, emptied records are counted\n\
    as records written.\n\
-f: Do not fail if the input file does not have a valid size, and do no\n\
    checking for utmp record validity.  This is liable to produce unuseful\n\
    output if the input becomes unsynchronized, since the type test will\n\
    likely be done on a field component that is not the type.\n\
-p<num-records>: Print a progress report to stderr after reading each\n\
    <num-records> records.\n\
-s<num-records>: Skip the initial <num-records> records in the file\n\
    without looking at them.  Note: the skipped records are not considered\n\
    deleted for the purpose of the -z and -O options; empty records are not\n\
    written in place of them.\n\
-w<num-records>: Write <num-records> and then quit.\n\
-o: Overwrite the input file.  Instead of writing records to the standard\n\
    output, they are written to the same file as the input.  Records are\n\
    actually only written after at least one record has been deleted,\n\
    since up to that point data identical to the file contents would be\n\
    written.  If -z is also given, the only records written are the emptied\n\
    records.  For the purpose of the -w option, records are considered to\n\
    have been written regardless of whether they were actually written or\n\
    the file pointer just moved past them.\n\
    After all records have been written, if the size of the file is\n\
    reduced due to deletion of records, skipping of bad records, or the\n\
    use of the -s or -w options, it is truncated so that it ends just\n\
    after the last record written.  The -o option can be used to avoid\n\
    using a large amount of extra disk space while compressing a big wtmp\n\
    file, but adds risk because if the compression is interrupted the file\n\
    is liable to be left with some records duplicated.\n\
-O: Like -o, except that any deleted records are immediately emptied.  This\n\
    uses more CPU time but is safer because if the compression is\n\
    interrupted, emptied records rather than duplicate records will be left\n\
    in the file.\n\
-n: Do not discard LOGIN_PROCESS, INIT_PROCESS, and EMPTY entries (unless\n\
    explicitly specified).  This can be used to fix a file with the -v or\n\
    -a option without compressing it.\n\
-d<name-prefix>: Discard entries whose ut_user field begins with\n\
    <name-prefix>.\n\
-i<ID[:ID...]>: Discard entries whose ID field is one of those given in\n\
    the colon-separated list.\n\
-T: Discard entries which have a non-empty user field and an empty line\n\
    field.  These are usually commands run from inittab which are not\n\
    associated with a tty.\n\
-0..-9: Remove entries with the given numeric type.  If any of these\n\
    options are given, entries of type LOGIN_PROCESS, INIT_PROCESS, and\n\
    EMPTY are not removed unless their flags are explicitly given.  See\n\
    utmp(F) for a description of the types.\n\
-u: This turns on the -n, -z, and -o options, and changes the default\n\
    file to %s"
#ifdef EXTENDED
" (or if -e is given, %s)"
#endif
".  It is used for\n\
    cleaning up the utmp file by zeroing (overwriting with empty) stale\n\
    entries, usually by using the -i option.\n\
-U: Like -u, except that records of type USER_PROCESS that have PIDs that\n\
    are no longer active and all records of type DEAD_PROCESS are\n\
    automatically emptied.  This will not necessarily clear all stale utmp\n\
    entries, because PIDs are reused.\n",
	    WTMP_FILE,
#ifdef EXTENDED
	    WTMPX_FILE, UTMPX_FILE,
#endif
	    UTMP_FILE
#ifdef EXTENDED
	    , UTMPX_FILE
#endif
	    );
	    exit(0);
	case 'd':
	    DeleteName = optarg;
	    DeleteLen = strlen(DeleteName);
	    break;
#ifdef EXTENDED
	case 'e':
	    Extended = 1;
	    formatSpecified = 1;
	    break;   
	case 'N':
	    Extended = 0;
	    formatSpecified = 1;
	    break;   
#endif
	case 'i':
	    if (!(IDlist = StrToArr(optarg,":",'\0'))) {
		fprintf(stderr,"cwtmp: bad ID string.\n");
		exit(1);
	    }
	    break;
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	    RemoveTypes[c - '0'] = 1;
	    TypesGiven = 1;
	    break;
	case 'n':
	    TypesGiven = 1;
	    break;
	case 'r':
	    Readable = 1;
	    break;
	case 'R':
	    Readable = 1;
	    verboseRecords = 1;
	    break;
	case 'H':
	    verboseRecords = 1;
	    break;
	case 'a':
	    utmpConfig(UTMP_AUTOVERIFY,NULL);
	    break;
	case 'f':
	    utmpConfig(UTMP_NOSIZECHECK,NULL);
	    break;
	case 's': case 'p': case 'w': case 'x':
	    if ((Value = atol(optarg)) < 1) {
		fprintf(stderr,
"cwtmp: argument to -%c must be a positive integer.  Use -h for help.\n",c);
		exit(1);
	    }
	    switch (c) {
	    case 's':
		Skip = Value;
		recordsSkipped = 1;
		break;
	    case 'p':
		utmpConfig(UTMP_PROGRESS,&Value);
		break;
	    case 'w':
		Write = Value;
		break;
	    case 'x':
		Debug = Value;
		fprintf(stderr,"Debugging set to level %d\n",Debug);
		utmpConfig(UTMP_VERBOSE,NULL);
		break;
	    }
	    break;
	case 'T':
	    DelInitCmds = 1;
	    break;
	case 't':
	    tell = 1;
	    break;
	case 'v':
	    utmpConfig(UTMP_VERIFY,NULL);
	    break;
	case 'O':
	    safeZero = 1;
	case 'o':
	    overwrite = 1;
	    /* Turn on indication of bad records so that we will know when
	    a section is skipped (turning on UTMP_RET_BAD causes an invalid
	    record type, which will always be deleted, to be returned for each
	    skipped section) */
	    utmpConfig(UTMP_RET_BAD,NULL);
	    break;
	case 'z':
	    zero = 1;
	    break;
	case 'U':
	    autoClean = 1;
	case 'u':
	    TypesGiven = zero = overwrite = 1;
	    utmpConfig(UTMP_RET_BAD,NULL);
	    uflag = 1;
	    break;
	case '?':
	    Err = 1;
	}
    if (Err)
	exit(1);
    if (Write && overwrite) {
	fputs("cwtmp: -w cannot be given with overwrite options.\n",stderr);
	exit(1);
    }
    if (Readable && overwrite) {
	fputs("cwtmp: -r cannot be given with overwrite options.\n",stderr);
	exit(1);
    }
    if (zero && safeZero) {
	fputs("cwtmp: -z cannot be given with -O.\n",stderr);
	exit(1);
    }
    /* This is implemented by doing only the safe-zeroing. */
    if (overwrite && zero) {
	zero = 0;
	safeZero = overZero = 1;
	/* Mark this fd as not being used, since only the emptyFD is used. */
	outFD = -1;
    }
    if (Debug) {
	if (safeZero)
	    fputs("Doing safe-zeroing\n",stderr);
    }
    if (Extended) {
	utmpConfig(UTMP_EXTENDED,NULL);
	ut_size = sizeof(struct utmpx);
    }

#ifdef UTMP_IS_UTMPX
    utmpConfig(UTMP_EXTENDED,NULL);
    Extended = 1;
    ut_size = sizeof(struct utmpx);
#endif

    switch (argc - optind) {
    case 0:	/* No input file specified */
#ifdef EXTENDED
	if (Extended)
	    if (uflag)
		InFile = UTMPX_FILE;
	    else
		InFile = WTMPX_FILE;
	else
#endif
	    if (uflag)
		InFile = UTMP_FILE;
	    else
		InFile = WTMP_FILE;
	break;
    case 1:
	InFile = argv[optind];
	break;
    default:
	fputs("cwtmp: too many arguments.  Use -h for help.\n",stderr);
	exit(1);
    }
    if (!TypesGiven) {
	RemoveTypes[EMPTY] = 1;
	RemoveTypes[LOGIN_PROCESS] = 1;
	RemoveTypes[INIT_PROCESS] = 1;
    }
    if (strcmp(InFile, "-") == 0)
	utmpConfig(UTMP_FD, 0);
    else
	utmpConfig(UTMP_NAME, InFile);
#ifdef EXTENDED
    if (!formatSpecified)
	utmpConfig(UTMP_AUTOSENSE,NULL);
#endif
    if ((orig_record = n_getutent(Skip)) == NULL) {	/* Open utmp file */
	if (n_getutent_err == NUTMP_FAIL_AUTO)
	    fprintf(stderr, "cwtmp: Use -e or -N to specify the file type.\n");
	exit(1);
    }
#ifdef EXTENDED
    if (!formatSpecified)
	if (Extended = utmpConfig(UTMP_TYPE,NULL))
	    ut_size = sizeof(struct utmpx);
#endif
    if (overwrite && !overZero && !tell)
	if ((outFD = open(InFile,O_WRONLY)) == -1) {
	    perror("cwtmp: Could not reopen input file for writing");
	    exit(1);
	}
	else if (Debug)
	    fprintf(stderr,"Reopened input file for output\n");
    if (!Readable && !tell && outFD != -1 && isatty(outFD)) {
	fputs(\
	"cwtmp: output is a tty; will not write raw utmp/wtmp records to it.\n"\
	"Use -h for help.\n",
	stderr);
	exit(2);
    }
    if (Debug)
	fprintf(stderr,"Output file is fd %d\n",outFD);
    if (tell)
	fputs("Would discard these records:\n",stdout);
    else if (safeZero) {
	if ((emptyFD = open(InFile,O_WRONLY)) == -1) {
	    perror(
	    "cwtmp: Could not reopen input file for writing zero-records");
	    exit(1);
	}
	if (isatty(emptyFD)) {
	    fputs(\
"cwtmp: Input is a tty; will not write raw (zero) utmp/wtmp records to it.\n",
	    stderr);
	    exit(2);
	}
	if (Debug)
	    fprintf(stderr,"Zero-record output file is fd %d\n",emptyFD);
    }
    if (Debug)
	fputs("Beginning file processing...\n",stderr);
    do {
	/* Remove record if its type is invalid or is in the list of types
	   to remove, or the user name begins with the prefix that was given
	   with -d, or its ID was given with -i, or if -t was given and
	   the entry has a non-empty user field and an empty line field,
	   or if autocleaning utmp & USER_PROCESS & dead PID */
	goodRec = 1;
	/*
	 * Generate a record that is always in utmpx format,
	 * for testing and printing
	 */
#ifdef EXTENDED
	if (Extended)
	    ux = orig_record;
	else
#endif
	    ux = u2ux(orig_record);
	if (ux->ut_type < 0 || ux->ut_type > UTMAXTYPE ||
	RemoveTypes[ux->ut_type] ||
	DeleteLen && !strncmp(ux->ut_user,DeleteName,DeleteLen) ||
	DelInitCmds && *ux->ut_user && !*ux->ut_line ||
	autoClean && (ux->ut_type == DEAD_PROCESS ||
	ux->ut_type == USER_PROCESS && !ProcExists(ux->ut_pid)))
	    goodRec = 0;
	else if (IDlist) {
	    for (id = IDlist; *id && strncmp(*id,ux->ut_id,sizeof(ux->ut_id)); id++)
		;
	    if (*id)
		goodRec = 0;
	}
	if (!goodRec) {
	    if (tell) {
		utmpxPrint(stdout,ux,verboseRecords,NULL,Extended);
		continue;
	    }
	    /* search_skipped is the number of bytes of bad utmp data skipped
	       by the nutmp routines.  If bad data was skipped, don't try to
	       zero it, since that would just make it harder to recognize
	       the bad section if we need to recover */
	    if (safeZero && !search_skipped) {
		/* If we are zeroing records, seek the empty-record writing FD
		   to the start of the record just read.
		   utmp_bytes_read is the absolute offset to the start of the
		   next utmp record after the one currently being processed.  */
		zeroWritePos = utmp_bytes_read - ut_size;
		if (emptyFDpos != zeroWritePos &&
		lseek(emptyFD,emptyFDpos = zeroWritePos,SEEK_SET) == -1) {
		    perror(
		    "cwtmp: lseek on overwrite file to do record-zero failed"
		    );
		    exit(1);
		}
		if (Debug > 2)
		    fprintf(stderr,"Writing zero record at record offset %lu\n",
			    emptyFDpos/ut_size);
		if ((numWrite = write(emptyFD,&emptyRec,ut_size)) != ut_size) {
		    if (numWrite == -1)
			perror("cwtmp: zero-record write failed");
		    else {
			fprintf(stderr,
		    "cwtmp: zero-record write only wrote %d bytes.  Exiting.\n",
			numWrite);
		    }
		    exit(1);
		}
		emptyFDpos += numWrite;
	    }
	    if (!recordsSkipped) {
		/* Start writing records */
		recordsSkipped = 1;
		/* Position write pointer past all of the records not-written
		   so far */
		if (overwrite && !overZero &&
		lseek(outFD,ut_size * writeRec,SEEK_SET) == -1) {
		    perror("cwtmp: lseek on output to first deleted record");
		    exit(1);
		}
	    }
	    /* zero is used only here */
	    if (zero && !search_skipped)
		orig_record = ux = &emptyRec;
	    else
		continue;
	}
	if (!tell) {
	    if (Readable)
		utmpxPrint(stdout,ux,verboseRecords,NULL,Extended);
	    else if ((!overwrite || recordsSkipped) && !overZero) {
		if (Debug > 2)
		    fprintf(stderr,"Writing record at record offset %d\n",
		    writeRec);
	        if ((numWrite = write(outFD,orig_record,ut_size))
			!= ut_size) {
		    if (numWrite == -1) {
			terrno = errno;
			fprintf(stderr,
			"cwtmp: Error writing to output at record %u: ",
			writeRec);
			errno = terrno;
			perror("");
		    }
		    else {
			fprintf(stderr,
		    "cwtmp: utmp record write only wrote %d bytes.  Exiting.\n",
			numWrite);
		    }
		    exit(1);
		}
	    }
	}
	if (Write && !--Write) {
	    recordsSkipped = 1;
	    break;
	}
	writeRec++;
    } while ((orig_record = n_getutent(0)) != NULL);
    if (!tell && overwrite && !overZero && recordsSkipped &&
	    ftruncate(outFD,ut_size * writeRec) == -1) {
	perror("cwtmp: truncate of overwrite file failed");
	exit(1);
    }
    exit(0);
}

int ProcExists(PID) {
    int ret;

    if ((ret = kill(PID,0)) == -1) {
	if (Debug > 2)
	    fprintf(stderr,"Kill of pid %d returned %d; errno = %d\n",
	    PID,ret,errno);
	return errno != ESRCH;
    }
    else {
	if (Debug > 2)
	    fprintf(stderr,"Kill of pid %d returned %d\n",PID,ret);
	return 1;
    }
}
