/*
*    dynarr.h
*/

/*
*    dynarr.h: dynamic array manipulation macros.
*    These macros create, assign to, resize,
*    and access arrays using allocated memory.
*    They are intended to simplify the use of arrays where elements are
*    added one at a time and the final array size is not initially known,
*    and the use of strcpy and strcat type operations on strings where the
*    final length is not known.
*
*    1989 john h. dubois iii (john@armory.com)
*    2010-10-24 Use stdlib.h.
*/
#ifdef _M_SYSV
#define M_SYS5
#endif
#include <stdlib.h>

struct AllocArr {    /* The dynamic array structure */
#ifdef M_SYS5
    void *Array;    /* Points to the dynamic array itself */
#else
    char *Array;
#endif
    /* Number of elements that are added to the array
       (via realloc()) when it is outgrown */
    size_t Inc;
    size_t CurElem; /* Which element of the array is to be assigned to next */
    size_t CurSize; /* How many elements are currently allocated to the array */
};

/*
*    InitArrRec: Initialize array record.
*    This macro sets the fields of the array record,
*    and does the initial memory allocation for the array.
*
*    struct AllocArr Rec: Array record.
*    Type is the type of the array element; i.e., char *, int, etc.
*    size_t Incr is the number of elements that should be added to the array
*    when it runs out of space.
*    size_t InitElems is the number of elements the array should initially
*    be malloced to.
*/
#define InitArrRec(Rec,Type,Incr,InitElems) \
((Rec).Array = (void *) calloc((InitElems),sizeof(Type)), \
(Rec).Inc = (Incr) * sizeof(Type), \
(Rec).CurElem = 0, \
(Rec).CurSize = (InitElems))

/*
*    AddElem: Add an element to a dynamic array.
*    This macro adds an element to a dynamic array,
*    realloc()ing it if necessary.
*
*    struct AllocArr Rec is the array record.
*    size_t Elem is the element to be added to the array.
*    Type is the type of the element.
*/
#define AddElem(Rec,Elem,Type) \
(((Type *) (((Rec).CurElem == (Rec).CurSize) ? \
((Rec).Array = (void *) \
realloc((Rec).Array,((Rec).CurSize += (Rec).Inc)*sizeof(Type)))\
: (Rec).Array)) [(Rec).CurElem++] = (void *) (Elem))

/*
*    SetElemNum: Set the current element number of an array record.
*    This macro sets the index of the next element of a dynamic array
*    that will be stored to.
*    It does not change the allocated size of the array.
*    The current element number should be set to a value less than or equal
*    to a value it has previously had.
*
*    struct AllocArr Rec is the array record.
*    size_t ElemNum is the value to set to element number to.
*/
#define SetElemNum(Rec,ElemNum) ((Rec).CurElem = (ElemNum))

/*
*    NumElem: get the current element number.
*    This macro returns the number of elements stored to a dynamic array,
*    which is also the index of the next element that will be stored to.
*
*    struct AllocArr Rec is the array record.
*/
#define NumElem(Rec) ((Rec).CurElem)

/*
*    ArrPt: get array pointer.
*    This macro returns the array from an array record.
*    The pointer becomes invalid with the next invocation of AddElem.
*
*    struct AllocArr Rec is the array record.
*/
#define ArrPt(Rec) ((Rec).Array)

/*
*     Dynamic string macros
*/

struct DynStr { 	/* Dynamic string structure */
    char *S;	    /* The string itself */
    size_t CurLen;	/* Current length of the string */
			/* (may be less than the size of the array) */
    size_t OldLen;
    size_t CurSize;     /* Current (malloced) size the char array */
};

/*
*    InitDynStr: Initialize string record.
*    This macro sets the fields of the string record,
*    and does the initial memory allocation for the string.
*
*    struct DynStr Rec: String record.
*    size_t InitElems: size the char array should initially be malloced to.
*/
#define InitDynStr(Rec,InitElems) \
((Rec).S = calloc((InitElems),sizeof(char)), \
(Rec).CurLen = 0, \
(Rec).OldLen = 0, \
(Rec).CurSize = (InitElems))

/*
*	dstrcpy: Copy a string into a dynamic string.
*	struct DynStr Rec: dynamic string to copy the string into.
*	char *Str: string to copy in.
*/
#define dstrcpy(Rec,Str) \
strcpy(((Rec).OldLen = (Rec).CurLen = strlen(Str)) >= (Rec).CurSize ? \
(Rec).S = realloc((Rec).S,(Rec).CurSize = (Rec).CurLen + 1) : (Rec).S,(Str))

/*
*	dstrcat: catenate a string to a dynamic string.
*	struct DynStr Rec: dynamic string to catenate the string to.
*	char *Str: string to catenate.
*/
#define dstrcat(Rec,Str) \
(strcpy((((Rec).CurLen += strlen(Str)) >= (Rec).CurSize ? \
(Rec).S = realloc((Rec).S,(Rec).CurSize = (Rec).CurLen + 1) : \
(Rec).S) + (Rec).OldLen,(Str)),(Rec).OldLen = (Rec).CurLen,(Rec).S)

/*
*	dincsize: change the memory allocated to a dynamic string
*	This macro does not change the CurLen field.
*	struct DynStr Rec: dynamic string to catenate the string to.
*	size_t Size: new size for the char array.
*/
#define dincsize(Rec,Size) \
((Size) > (Rec).CurSize ? \
(Rec).S = realloc((Rec).S,(Rec).CurSize = (Size)) : (Rec).S)
