/*
 *  nutmp.c: replacement for standard library utmp routines, which have poor
 *  error reporting and refuse to operate on a corrupt (wrong size) file.
 *  1994-12-12 John H. DuBois III (john@armory.com)
 *  1995-07-23 Added Progress and skip
 *  1996-04-13 Require ut_time to be < current_time+7200
 *  1996-12-24 Added UTMP_RET_BAD feature.
 *             Allow ut_time to be 0 without failing good-time test.
 *  2000-01-12 Added verbose & time_format flags to utmpPrint()
 *  2000-10-11 Made time test work properly.
 *  2001-07-19 Added utmpx support.
 *  2003-01-15 Be more careful about using ut_host, in case it is not null
 *             terminated.
 *             Linux & UnixWare port.
 *             Use UT_*SIZE macros.
 *  2005-05-07 Added utmp file type autosense.
 *  2008-03-04 Make openUtmp() return various error values
 *             Make n_getutent set n_getutent_err
 *             Detect empty files
 *  2010-10-24 Linux re-port/64-bit port.
 *             #include a few more headers, use new method to get MAXPID,
 *             avoid problems due to ut_tv/timeval difference, various cleanup.
 */

/*
 * This code currently only works for systems that have a utmpx structure;
 * will have to be cleaned up to work again on those that don't.
 * The Linux utmp struct is equivalent to the SYSV utmpx struct,
 * but without ut_syslen and with the additional element int32_t ut_addr_v6[4],
 * which we currently maintain but otherwise don't do anything with.
 */
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/param.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include "nutmp.h"

#if defined(UTMP_IS_UTMPX) && !defined(EXTENDED)
#define EXTENDED
#endif

/* 
 * These are defined under Linux.
 * If not defined, set them to the sizes of the utmpx structure elements.
 */

#ifndef UT_LINESIZE
#define UT_LINESIZE ((int)(sizeof((struct utmpx *)0)->ut_line))
#endif
#ifndef UT_NAMESIZE
#define UT_NAMESIZE ((int)(sizeof((struct utmpx *)0)->ut_user))
#endif
#ifndef UT_HOSTSIZE
#define UT_HOSTSIZE ((int)(sizeof((struct utmpx *)0)->ut_host))
#endif

#define UT_IDSIZE	((int)(sizeof((struct utmpx *)0)->ut_id))

#define NO_UTSYSLEN UTMP_IS_UTMPX

static FILE *utmpFile = NULL;		/* utmp file being read */
static char *_utmpName = UTMP_FILE;	/* Name of utmp file to read */
static int utmp_fd = -1;		/* fd to read from */
static int Verify;			/* Check records for sanity */
static int AutoVerify;	/* Check records if size of utmp file is not correct */
static int Verbose;			/* Print per-record info */
static int SizeFail = 1;	/* Abort if size of utmp file is not correct */
static long RecordNum;			/* Number of records read */
static long SkippedBytes;		/* Number of bytes skipped */
/* Number of records after which to print progress indication */
static long Progress = 0;
/* Number of records before next progress indication should be printed */
static long ProgressCt;
/* Return an indication for each bad record rather than skipping it */
static int returnBad = 0;
static int utsize = sizeof(struct utmp);
static char *utstructName = "utmp";	/* For error messages */
static int Extended = 0;		/* Processing a utmpx file */
static pid_t maxPID = 0;

/* This is made into a special record & returned to indicate a bad record */
static struct utmp badRec;
#ifdef EXTENDED
static struct utmpx badRecx;
static int Autosense = 0;		/* Automatic file type detection */
#endif

/* Prototypes for local functions */
static char *sTest(char *s, int n);
static int timeTest(int sec, int usec);
static pid_t getMaxPID(void);
static int utmpTest(struct utmpx *u);
static unsigned _nutmp_strnlen(const char *s, size_t n);

/*
 * Configure library.  See declaration of variables modifed here for
 * explanation of what these calls do.
 */
int
utmpConfig(int a, void *Arg)
{
    switch (a) {
    case UTMP_RET_BAD:
	/*
	 * Note that UTMP_RET_BAD does not turn on Verify; unless UTMP_VERIFY
	 * is done separately UTMP_RET_BAD will have no effect
	 */
	returnBad = 1;
	break;
    case UTMP_VERBOSE:
	Verbose = 1;
	break;
    case UTMP_PROGRESS:
	Progress = ProgressCt = *((long *)Arg);
	break;
    case UTMP_VERIFY:
	Verify = 1;
	SizeFail = 0;
	break;
    case UTMP_NOSIZECHECK:
	SizeFail = 0;
	break;
#ifdef EXTENDED
    case UTMP_EXTENDED:
	Extended = 1;
	utsize = sizeof(struct utmpx);
	utstructName = "utmpx";
	break;
    case UTMP_AUTOSENSE:
	Autosense = 1;
	utsize = 0;
	utstructName = NULL;
	break;
    case UTMP_TYPE:
	return Extended;
	break;
#endif
    case UTMP_REG:
	Extended = 0;
	utsize = sizeof(struct utmp);
	utstructName = "utmp";
	break;
    case UTMP_AUTOVERIFY:
	AutoVerify = 1;
	SizeFail = 0;
	break;
    case UTMP_NAME:
	/*
	 * Change the name of the file read by the utmp routines.
	 */
	if (!(_utmpName = strdup(Arg)))
	    return 2;
	utmp_fd = -1;
	break;
    case UTMP_FD:
	utmp_fd = *(int *)Arg;
	_utmpName = "(input)";
	break;
    default:
	return 1;
    }
    return 0;
}

/*
 * This will always give the number of bytes read, which will normally be
 * the offset of the next byte after the last one in the record being returned.
 * If an indication of a bad region skipped is being returned, its value is
 * undefined outside of this module.
 */
long utmp_bytes_read = 0;

/*
 * Open a utmp[x] file.
 * Return value:
 * On success, 0.
 * On failure, a nonzero value:
 * NUTMP_FAIL_OPEN if the file cannot be opened
 * NUTMP_FAIL_READ if the file cannot be read
 * NUTMP_FAIL_AUTO if automatic type detection cannot be used
 * NUTMP_FAIL_TYPE if the file does not contain records of the expected type
 * NUTMP_FAIL_SIZE if the file has a size not a multiple of the type of structure it should
 *   contain
 * NUTMP_FAIL_EMPTY if the file does not contain any non-null records
 */
static int
openUtmp()
{
    struct stat statbuf;
    /* Record read for automatic type detection */
    static struct utmpx pushedRecord;

    /*
     * It's OK for this to not be a multiple of record size because it's just
     * used as a stdio buffer, so no adjustment for utmpx records
     */
#define INBUFSIZE (sizeof(struct utmp) * 2000)
    static char infileBuf[INBUFSIZE];
    long offset;

    if (maxPID == 0)
	maxPID = getMaxPID();

    if (utmp_fd == -1)
	utmpFile = fopen(_utmpName,"r");
    else
	utmpFile = fdopen(utmp_fd,"r");
    if (utmpFile == NULL) {
	fprintf(stderr,"Could not open file \"%s\" for reading: \n",_utmpName);
	fflush(stderr);
	perror("");
	return NUTMP_FAIL_OPEN;
    }
    /* Let this fail silently */
    setvbuf(utmpFile, infileBuf, _IOFBF, INBUFSIZE);
    utmp_bytes_read = RecordNum = 0;
#ifdef EXTENDED
    if (Autosense) {
	/* We will need to rewind, because testing the file type may require
	 * skipping over an indefinite number of empty records.
	 */
	offset = ftell(utmpFile);
	if (fseek(utmpFile, offset, SEEK_SET)) {
	    fprintf(stderr,"Cannot use automatic utmp file type detection on a non-seekable device\n");
	    fflush(stderr);
	    fclose(utmpFile);
	    return NUTMP_FAIL_AUTO;
	}
	switch (testFileRead(utmpFile, offset, 0)) {
	case 1:
	    return NUTMP_FAIL_EMPTY;
	case 2:
	    return NUTMP_FAIL_READ;
	case 3:		/* Not a regular utmp file */
	    switch (testFileRead(utmpFile, offset, 1)) {
	    case 1:
		return NUTMP_FAIL_EMPTY;	/* shouldn't happen */
	    case 2:
		return NUTMP_FAIL_READ;
	    case 3:	/* Not an extended utmp file */
		fprintf(stderr,"Could not determine type of file \"%s\".\n",_utmpName);
		fflush(stderr);
		fclose(utmpFile);
		return NUTMP_FAIL_TYPE;
	    }
	}
    }
    if (Verbose)
	fprintf(stderr,"%s is a %s file\n", _utmpName, utstructName);
    /* Must do this after autosense to get utmp struct size */
    if (!stat(_utmpName, &statbuf) && statbuf.st_size % utsize) {
	fprintf(stderr,
	"%s: size of %s (%ld bytes) is not a multiple of the size of a\n"\
	"%s record (%d bytes).\n",SizeFail ? "Error" : "Warning",
	_utmpName,statbuf.st_size,utstructName,utsize);
	if (SizeFail)
	    return NUTMP_FAIL_SIZE;
	if (AutoVerify)
	    Verify = 1;
    }
    if (Verbose)
	fprintf(stderr,"Successfully opened %s file %s\n",utstructName,
	_utmpName);
    if (Extended)
	badRecx.ut_type = UTMP_BAD;
    else
#endif
	badRec.ut_type = UTMP_BAD;
    return 0;
}

#ifdef EXTENDED
/*
 * Return value:
 * 0 if utmpFile contains records of the expected type
 * 1 if utmpFile does not contain any non-null records after offset
 * 2 if attempting to read from utmpFile gives an error
 * 3 if utmpFile contains data that does not consist of records of the expected
 *   type
 */
int
testFileRead(FILE *utmpFile, long offset, int tryExtended)
{
    size_t ret;
    struct utmpx record;
    int ateof;

    /*
     * Skip over empty records, which are likely to occur at the start of
     * utmp/utmpx, and provide no information for testing against
     */
    while ((ret = fread(&record, tryExtended ? sizeof(struct utmpx) : sizeof(struct utmp), 1, utmpFile)) == 1 &&
	    (tryExtended ? record.ut_type : ((struct utmp *)&record)->ut_type) == EMPTY)
	;
    ateof = feof(utmpFile);	/* must get this before fseek */
    fseek(utmpFile, offset, SEEK_SET);
    if (ret != 1)
	if (ateof) {
	    if (offset == 0) {
		fprintf(stderr,"No non-empty records in file \"%s\"\n", _utmpName);
		fflush(stderr);
	    }
	    return 1;
	}
	else {
	    fprintf(stderr,"Could not read from file \"%s\":\n", _utmpName);
	    fflush(stderr);
	    perror("");
	    return 2;
	}
    utmpConfig(tryExtended ? UTMP_EXTENDED : UTMP_REG, NULL);
    return utmpTest(tryExtended ? &record : u2ux((struct utmp *) &record)) ? 0 : 3;
}
#endif

/* This must be updated if new types are added */
static char *v_types[] = {
    "EMPTY", "RUN_LVL", "BOOT_TIME", "OLD_TIME", "NEW_TIME", "INIT_PROCESS",
    "LOGIN_PROCESS", "USER_PROCESS", "DEAD_PROCESS", "ACCOUNTING"
};

/*
 * Convert a utmp record to a utmpx record
 */
struct utmpx *
u2ux(struct utmp *u)
{
#ifdef UTMP_IS_UTMPX
    return (struct utmpx *)u;
#else
    static struct utmpx ux;

    /*
     * Ensure that any utmpx string elements that are larger than the 
     * equivalent utmp element are null terminated after the copied part.
     */
    bzero(&ux,sizeof(struct utmpx));
    strncpy(ux.ut_user,u->ut_user,sizeof((struct utmp *)0)->ut_user);
    strncpy(ux.ut_id,u->ut_id,(sizeof((struct utmp *)0)->ut_id));
    strncpy(ux.ut_line,u->ut_line,(sizeof((struct utmp *)0)->ut_line));
    ux.ut_pid = u->ut_pid;
    ux.ut_type = u->ut_type;
    ux.ut_tv.tv_sec = u->ut_time;
    ux.ut_tv.tv_usec = 0;
    ux.ut_exit.e_termination = u->ut_exit.e_termination;
    ux.ut_exit.e_exit = u->ut_exit.e_exit;
    ux.ut_syslen = 1;	/* just the terminating null char */
    return &ux;
#endif
}

void
utmpxPrint(FILE *f, struct utmpx *u, int verbose, char *time_format, int printExt)
{
    char buf[81];

    fprintf(f, "u=%-8.*s ", UT_NAMESIZE, u->ut_user);
    fprintf(f, "id=%-*.*s ", UT_IDSIZE, UT_IDSIZE, u->ut_id);
    fprintf(f, "l=%-*.*s ", UT_LINESIZE < 12 ? UT_LINESIZE : 12, UT_LINESIZE, u->ut_line);
    fprintf(f, "p=%-5u ", u->ut_pid);
    /*
     * this tends to be garbage
     * if (printExt)
     *     fprintf(f,"ses=%-5u ",u->ut_session);
     */
    if (verbose) {
	struct timeval tv;

	if (u->ut_type <= 9)
	    fprintf(f,"%-9.9s ",v_types[u->ut_type]);
	else
	    fprintf(f,"type=%-4u ",(unsigned) u->ut_type);
	/* struct timeval and ut_tv may not be identical */
	tv.tv_sec = u->ut_tv.tv_sec;
	tv.tv_usec = u->ut_tv.tv_usec;
	strftime(buf,sizeof(buf)-1,time_format ? time_format : "%Y-%m-%d %T %Z",
		localtime((time_t *) &tv));
	fprintf(f,"%s ",buf);
    }
    else {
	fprintf(f,"type=%-1u ",(unsigned) u->ut_type);
	if (printExt) {
	    sprintf(buf,"%ld.%ld",(long)u->ut_tv.tv_sec, (long)u->ut_tv.tv_usec);
	    fprintf(f,"time=%-17s ",buf);
	}
	else
	    fprintf(f,"time=%-10ld ",(long) u->ut_tv.tv_sec);
    }
    fprintf(f,"trm=%-5hu ",u->ut_exit.e_termination);
    fprintf(f,"ex=%-3hu ",u->ut_exit.e_exit);
    if (printExt && u->ut_host[0])
	fprintf(f,"%.*s",
#ifdef UTSYSLEN
		u->ut_syslen
#else
		UT_HOSTSIZE
#endif
		,u->ut_host);
    fputc('\n',f);
}

/*
 * Search buffer mem of length n for a character that is not c.
 * If such a character is found, a pointer to it is returned.
 * If not, a null pointer is returned.
 */
static char *
notchar(char *mem, char c, int n)
{
    char *orig = mem;

    for (; n; n--) {
	if (*mem != c) {
	    return mem;
	}
	mem++;
    }
    return (char *) 0;
}

/*
 * Returns true if u is a good utmpx structure, according to these tests:
 * utmpx struct elements:
 *	char ut_user		0 or more printable chars followed by nulls
 *	char ut_id		0 or more printable chars followed by nulls
 *	char ut_line		0 or more printable chars followed by nulls
 *	pid_t ut_pid		Integer 0..MAXPID
 *	short ut_type		Integer 0..UTMAXTYPE
 *				Also, if type is 0, entire record must be 0
 *	struct ut_exit		0 <= e_termination <= 65535 (not actually
 *	                        tested, since e_termination is a short),
 *				0 <= e_exit <= 255
 *	time_t ut_tv.tv_sec	Integer > 0, <= approx-current-time+7200
 *	time_t ut_tv.tv_usec	Integer in the range 0..999999
 *	long ut_session		no test
 *	char ut_host[]		0 or more printable chars followed by nulls.
 *	short ut_syslen		Must be >=0 and < size of ut_host, and must
 *				equal length of ut_host or ut_host+1
 *				(both seem to be used!)
*/
static int
utmpTest(struct utmpx *u)
{
    char *p;

    size_t hostlen;

    p = memchr(u->ut_host,'\0',UT_HOSTSIZE); 
    hostlen = p ? (p - u->ut_host) : 257;
	
    /*
    if (!(!u->ut_type && notchar((char *)u,'\0',utsize)) &&
    */
    if (
    !sTest(u->ut_user,UT_NAMESIZE) &&
    !sTest(u->ut_id,UT_IDSIZE) && !sTest(u->ut_line,UT_LINESIZE)
    && 0 <= u->ut_pid && u->ut_pid <= maxPID &&
    0 <= u->ut_type && u->ut_type <= UTMAXTYPE &&
    0 <= u->ut_exit.e_exit && u->ut_exit.e_exit <= 255 &&
    timeTest(u->ut_tv.tv_sec, u->ut_tv.tv_usec) &&
#ifdef UTSYSLEN
    0 <= u->ut_syslen && u->ut_syslen < UT_HOSTSIZE &&
    (u->ut_syslen == hostlen || u->ut_syslen-1 == hostlen) &&
#endif
    !sTest(u->ut_host,UT_HOSTSIZE)) {
	if (Verbose) {
	    if (Progress)
		fputc('\n',stderr);
	    fputs("Good record: ",stderr);
	    fprintf(stderr, 
"User name: %u chars; ID: %u chars; line: %u chars; PID = %d (<%d);\n\
type = %d (<%d), exit = %d [0..255], time = %d.%06d (0..%ld)\n",
		    _nutmp_strnlen(u->ut_user, sizeof(UT_NAMESIZE)),
		    _nutmp_strnlen(u->ut_id, sizeof(UT_IDSIZE)),
		    _nutmp_strnlen(u->ut_line, sizeof(UT_LINESIZE)),
		    u->ut_pid, maxPID, u->ut_type, UTMAXTYPE,
		    u->ut_exit.e_exit, u->ut_tv.tv_sec, u->ut_tv.tv_usec,
		    (long)(time(NULL) + 7200));
	    /*
	     * Should add prints of this stuff
#ifdef UTSYSLEN
	    if (!(0 <= u->ut_syslen && u->ut_syslen < UT_HOSTSIZE))
		fprintf(stderr,"syslen (%u); ",u->ut_syslen);
	    else if (u->ut_syslen != hostlen && u->ut_syslen-1 != hostlen)
		fprintf(stderr,"syslen (%u) != hostlen (%u) (+ 1); ",
			u->ut_syslen, hostlen);
	    else 
#endif
	    */
	    if (p = sTest(u->ut_host,UT_HOSTSIZE))
		fprintf(stderr,"host (%s); ",p);
	    fputc('\n',stderr);
	    utmpxPrint(stderr,u,0,NULL,Extended);
	}
	return 1;
    }
    else {
	if (Verbose) {
	    if (Progress)
		fputc('\n',stderr);
	    fputs("Bad: ",stderr);
	    /*
	    if (!u->ut_type && notchar((char *)u,'\0',utsize))
		fprintf(stderr,"type = empty but record not empty; ");
	    */
	    if (p = sTest(u->ut_user,UT_NAMESIZE))
		fprintf(stderr,"user (%s); ",p);
	    if (p = sTest(u->ut_id,UT_IDSIZE))
		fprintf(stderr,"id (%s); ",p);
	    if (p = sTest(u->ut_line,UT_LINESIZE))
		fprintf(stderr,"line (%s); ",p);
	    if (!(0 <= u->ut_pid && u->ut_pid <= maxPID))
		fputs("pid; ",stderr);
	    if (!(0 <= u->ut_type && u->ut_type <= UTMAXTYPE))
		fputs("type; ",stderr);
	    if (!(0 <= u->ut_exit.e_exit && u->ut_exit.e_exit <= 255))
		fputs("exit; ",stderr);
	    if (!timeTest(u->ut_tv.tv_sec, u->ut_tv.tv_usec))
		fputs("time; ",stderr);
#ifdef UTSYSLEN
	    if (!(0 <= u->ut_syslen && u->ut_syslen < UT_HOSTSIZE))
		fprintf(stderr,"syslen (%u); ",u->ut_syslen);
	    else if (u->ut_syslen != hostlen && u->ut_syslen-1 != hostlen)
		fprintf(stderr,"syslen (%u) != hostlen (%u) (+ 1); ",
			u->ut_syslen, hostlen);
	    else 
#endif
	    if (p = sTest(u->ut_host,UT_HOSTSIZE))
		fprintf(stderr,"host (%s); ",p);
	    fputc('\n',stderr);
	    utmpxPrint(stderr,u,0,NULL,Extended);
	}
	return 0;
    }
}

off_t search_skipped = 0;

/*
 * Read one character at a time, searching for the start of a
 * utmp[x] record.
 * Input variables:
 * badRec is the bad record found that prompted this search.  It should be
 * passed because the next good record may start somewhere in it.
 * Return value:
 * On success, a pointer to a static utmp[x] buffer containing the next good
 * record found.  On failure (no good record found before EOF), a null pointer.
 */
static void *
utmpSearch(void *badRec)
{
    int c;		/* Character read */
    char *p;		/* Buffer scanner */
    long StartOffset = utmp_bytes_read;
    struct utmpx *uxp;
    /*
     * Buffer[] is used as a circular buffer while searching for the start of a
     * record
     */
#define UBUFSIZE 10
    struct utmpx Buffer[UBUFSIZE];
    static struct utmpx retBuf;

    fprintf(stderr,
    "%s%s read out of sync at record %ld (byte offset %lu); searching...\n",
    Progress ? "\n" : "", utstructName, RecordNum, utmp_bytes_read);

    memcpy(Buffer,badRec,utsize);
    p = (char *) Buffer;	/* so we can do 1-byte increments */
    while ((c = fgetc(utmpFile)) != EOF) {
	utmp_bytes_read++;
	/* If we've reached the end of the buffer, copy the current structure
	   back to the beginning and point p there */
	if (p == (char *) (Buffer + (UBUFSIZE - 1))) {
	    memcpy(Buffer,p,utsize);
	    p = (char *) Buffer;
	}
	*(p + utsize) = (char) c;
	p++;
	/* p now points to a utmp[x]-size buffer of characters to test */
	if (Extended)
	    uxp = (struct utmpx *)p;
	else
	    uxp = u2ux((struct utmp *)p);
	if (utmpTest(uxp)) {
	    /* Found a valid record */
	    fprintf(stderr,
		    "%s read back in sync; skipped %lu bytes.  Sync record:\n",
		    utstructName, search_skipped = utmp_bytes_read - StartOffset);
	    SkippedBytes += search_skipped;
	    if (!returnBad)
		search_skipped = 0;
	    utmpxPrint(stderr,uxp,0,NULL,Extended);
	    memcpy(&retBuf,p,utsize);
	    return p;
	}
    }
    fprintf(stderr,
    "%s read failed to sync; at EOF after skipping %lu bytes.\n",
    utstructName,search_skipped = utmp_bytes_read - StartOffset);
    SkippedBytes += search_skipped;
    if (!returnBad)
	search_skipped = 0;
    return NULL;
}

int n_getutent_err = 0;

/*
 * Skip NumSkip records before reading the next record.
 * Return value: pointer to the next utmp[x] record on success;
 * NULL on failure (including EOF).
 * If NULL is returned, the global n_getutent_err is set to:
 *     0 if EOF
 *     The return value from openUtmp() if an open error occurs
 */
void *
n_getutent(long NumSkip)
{
    void *Ret;
    static struct utmpx u;
    /*
     * If a bad extent is found followed by a good record, we may want to
     * return a bad record as an indicator but return the good record next time
     * around.  In this case, nextRec will be set to the good record we should
     * return next time.
     */
    static void *nextRec = NULL;

    if (nextRec) {	/* If we have a saved good record from earlier */
	Ret = nextRec;
	nextRec = NULL;
	/*
	 * Must zero this because a "real" record with the (invalid) type
	 *  UTMP_BAD is indicated by the return of a record of that type
	 *  with search_skipped set to 0.  search_skipped is only set non-0
	 *  if returnBad is true, which ensures that nextRec will be set
	 *  so we are guaranteed to be able to clean up here.
	 */
	search_skipped = 0;
	return Ret;
    }
    if (!utmpFile && (n_getutent_err = openUtmp()) != 0)
	return NULL;	/* File open failed */
    if (NumSkip)
	if (fseek(utmpFile,NumSkip*utsize,SEEK_CUR)) {
	    if (Progress)
		fputc('\n',stderr);
	    fprintf(stderr,"Skip of %ld records in file \"%s\" failed.\n",
	    NumSkip,_utmpName);
	}
	else {
	    utmp_bytes_read += NumSkip * utsize;
	    RecordNum += NumSkip;
	    SkippedBytes += NumSkip * utsize;
	}
    if (fread(&u,utsize,1,utmpFile) == 1) {	/* Read next utmp[x] record */
	if (Verify && !utmpTest(Extended ? &u : u2ux((struct utmp *)&u))) {
	    /* If bad record... */
	    Ret = utmpSearch(&u);		/* Search for next good record */
	    /*
	     * If we should return a bad record as an indication when we
	     * encounter a bad extent, save the good record found (if any)
	     * for next time and set the record pointer to the special bad
	     * record.
	     */
	    if (returnBad) {
		nextRec = Ret;
		Ret = &badRec;
	    }
	}
	else
	    Ret = &u;
    }
    else
	Ret = 0;
    utmp_bytes_read += utsize;
    RecordNum++;
    if (!Ret) {	/* At EOF */
	if (Progress)
	    fputc('\n',stderr);
	if (SkippedBytes)
	    fprintf(stderr,"Skipped %lu bytes total in file %s\n",SkippedBytes,
	    _utmpName);
	n_getutent_err = 0;
	return NULL;
    }
    if (Progress && !--ProgressCt) {
	fprintf(stderr,"%ld records\r",RecordNum);
	ProgressCt = Progress;
    }
    return Ret;
}

/*****************************
 * non-utmp-specific functions
 ****************************/

/*
 * s is a character buffer.
 * n is the size of the buffer.
 * Return value:
 * If s consists of 0 or more printable chars followed by 0 or more nulls,
 * a null pointer is returned.
 * Otherwise, returns a static buffer containing a description of the problem.
 */
static char *
sTest(char *s, int n)
{
    static char buf[70];
    unsigned char *p = (unsigned char *)s;
    int gotNull;

    while (n-- && isprint(*(p++)))
	;
    if (n == -1)
	return NULL;	/* All printable chars */
    n++;
    p--;
    /* p now points to a non-printing char (possibly a null) */
    gotNull = !*p;
    if (gotNull)	/* Remove this to be less lenient */
	return NULL;
    /* Test if all chars after printable chars are nulls */
    while (n-- && !*(p++))
	;
    p--;
    /*
     * p now either points to a null char at the end of the buffer,
     * or points to a non-null char that occurred before the end of the buffer.
     */
    if (*p) {
	if (gotNull)
	    sprintf(buf,"Non-null char 0x%02x after null char",*p);
	else
	    sprintf(buf,"Non-printing char 0x%02x",*p);
	return buf;
    }
    else
	return NULL;
}

/*
 * Returns true if Time is a UNIX epoch time from 0 through the current time +
 * 2 hours.
 */
static int
timeTest(int sec, int usec)
{
    static time_t TestTime = (time_t) 0;

    /*
     * If time fails TestTime test, get time again in case it's just
     * because we've been processing a long time.
     * Actually, TestTime is initialized to 0 and only set here.
     */
    return (0 <= usec && usec < 1000000 && sec >= 0 &&
	    (sec <= TestTime || sec <= (TestTime = time(NULL) + 7200)) );
}

/*
 * _nutmp_strnlen: Find the length of a possibly null-terminated string,
 * with an upper bound.
 *
 * Input variables:
 * s is the string to find the length of.
 * n is the maximum length of the string.
 *
 * Return value:
 * The length of s.  No more than the first n characters of s are examined;
 * if a null character is not found within the first n characters, the length
 * of s is assumed to be n, and n is returned.
 */
static unsigned
_nutmp_strnlen(const char *s, size_t n) {
    const char *orig_s = s;

    while (n && *s) {
	s++;
	n--;
    }
    return s - orig_s;
}

#if !defined(MAXPID) && defined(PID_MAX)
#   define MAXPID PID_MAX
#endif

static pid_t
getMaxPID(void) {
    pid_t maxPID;
#ifdef MAXPID
    maxPID = MAXPID;
#elif defined(__linux__)
#   define PIDLEN 30
    unsigned char buf[PIDLEN+1];
    int fd;
    buf[PIDLEN] = '\0';
    if ((fd = open("/proc/sys/kernel/pid_max", O_RDONLY)) < 0 ||
	    read(fd, PIDLEN, buf) <= 1 || (maxPID = atol(buf)) < 1)
	maxPID = INT_MAX;
#else
    maxPID = INT_MAX;
#endif
    return maxPID;
}

